#include <Wire.h>
#include "RTClib.h"
#include <Adafruit_NeoPixel.h>
#ifdef __AVR__
#include <avr/power.h>
#endif

// Analog
#define PIN_PHOTORESISTOR  0

// Digital
#define PIN_ENCODER_A      3
#define PIN_ENCODER_SWITCH 4
#define PIN_ENCODER_B      5
#define PIN_PIXEL_DATA     6

// Used for direct port access
#define PORT_D_INPUTS PIND

#define SEGMENTS_PER_DIGIT 7
#define LEDS_PER_DIGIT 7
#define GRID_WIDTH 3
#define GRID_HEIGHT 5

#define _ 999

const byte NOOP_MODE = 0;
const byte SET_HOUR_MODE = 1;
const byte SET_MINUTE_MODE = 2;
const byte MODE_COUNT = 3;

const byte ANIMATION_STATE_NONE = 0;
const byte ANIMATION_STATE_FADING = 1;
const byte ANIMATION_STATE_GROWING = 2;
const byte ANIMATION_STATE_COUNT = 3;

const int DIGIT_SEGMENTS[11][SEGMENTS_PER_DIGIT] = {
  {1, 1, 1, 1, 1, 1, 0}, // 0
  {0, 0, 1, 1, 0, 0, 0}, // 1
  {0, 1, 1, 0, 1, 1, 1}, // 2
  {0, 1, 1, 1, 1, 0, 1}, // 3
  {1, 0, 1, 1, 0, 0, 1}, // 4
  {1, 1, 0, 1, 1, 0, 1}, // 5
  {1, 1, 0, 1, 1, 1, 1}, // 6
  {0, 1, 1, 1, 0, 0, 0}, // 7
  {1, 1, 1, 1, 1, 1, 1}, // 8
  {1, 1, 1, 1, 0, 0, 1}, // 9
  {0, 0, 0, 0, 0, 0, 0} // Digit off
};

RTC_DS3231 rtc;

Adafruit_NeoPixel strip = Adafruit_NeoPixel(30, PIN_PIXEL_DATA, NEO_GRB + NEO_KHZ800);

static bool updatedAppearance = false;
static uint8_t encPrevPos = 0;
static char encSwitchWasPressed = 0;
static byte mode = 0;
static byte prevMode = -1;
static byte brightness = 128;

static int prevHour = -1;
static int prevMinute = -1;

static bool animating = false;
static byte animationState = ANIMATION_STATE_NONE;

static unsigned long animationStartMillis = 0;
static unsigned long animationEndMillis = 0;

//
// SETUP
//
void setup() {
  Serial.begin(9600);
  setupPins();
  setupRtc();
  setupEncoder();
  setupNeoPixels();
}

void setupPins() {
  pinMode(PIN_ENCODER_A, INPUT_PULLUP);
  pinMode(PIN_ENCODER_B, INPUT_PULLUP);
  pinMode(PIN_ENCODER_SWITCH, INPUT_PULLUP);
}

void setupRtc() {
  if (!rtc.begin()) {
    Serial.println("Couldn't find RTC");
    while (1);
  }

  if (rtc.lostPower()) {
    Serial.println("RTC lost power, setting the time to the sketch compilation time.");
    rtc.adjust(DateTime(F(__DATE__), F(__TIME__)));
    // This line sets the RTC with an explicit date & time, for example to set
    // January 21, 2014 at 3am you would call:
    // rtc.adjust(DateTime(2014, 1, 21, 3, 0, 0));
  }
}

void setupEncoder() {
  // get an initial reading on the encoder pins
  if (digitalRead(PIN_ENCODER_A) == LOW) {
    encPrevPos |= (1 << 0);
  }
  if (digitalRead(PIN_ENCODER_B) == LOW) {
    encPrevPos |= (1 << 1);
  }
}

void setupNeoPixels() {
  strip.begin();
  strip.setBrightness(brightness);
  strip.show(); // Initialize all pixels to 'off'
}

//
// LOOP
//
void loop() {
  // Mode switch always implies that we updated the appearance
  updatedAppearance = mode != prevMode;
  prevMode = mode;

  DateTime now = rtc.now();
  unsigned int hour = now.hour();
  unsigned int minute = now.minute();

  unsigned long nowMillis = millis();
  
  if (prevMinute != minute) {
    prevMinute = minute;
    updatedAppearance = true;
  }
  
  int8_t encRotation = getEncoderRotation();
  if (encRotation != 0) {
    switch (mode) {
      case SET_HOUR_MODE:
        hour = (hour + 24 + encRotation) % 24;
        rtc.adjust(DateTime(now.year(), now.month(), now.day(), hour, now.minute(), now.second()));
        updatedAppearance = true;
        break;
      case SET_MINUTE_MODE:
        minute = (minute+ 60 + encRotation) % 60;
        rtc.adjust(DateTime(now.year(), now.month(), now.day(), now.hour(), minute, 0));
        updatedAppearance = true;
        break;
      default: // NOOP_MODE
        break;
    }
  }

  handleEncoderButton();

  adjustBrightness();

  if (animating || updatedAppearance) {
    uint32_t unixTime = now.unixtime();
    uint32_t daySeconds = unixTime % 86400;
    
    int colorCounter = map(daySeconds, 0, 86399, 0, 255);
  
    uint32_t color = wheel(colorCounter);
    uint32_t hourColor = mode == SET_HOUR_MODE ? calculateAnimatedColor(nowMillis, color) : color;
    uint32_t minuteColor = mode == SET_MINUTE_MODE ? calculateAnimatedColor(nowMillis, color) : color;

    // Adjust for 12 hour display instead of 24 hour
    if (hour > 12) {
      hour -= 12;
    } else if (hour == 0) {
      hour = 12;
    }

    if (hour < 10) {
      // Set the first digit to be blank (a value of 10)
      setDigitAt(0, 10, hourColor);
      setDigitAt(1, hour, hourColor);
    } else {
      setDigitAt(0, hour / 10, hourColor);
      setDigitAt(1, hour % 10, hourColor);
    }

    // Colon is always on and always a static color
    strip.setPixelColor(14, color);
    strip.setPixelColor(15, color);

    setDigitAt(2, minute / 10, minuteColor);
    setDigitAt(3, minute % 10, minuteColor);
  }

  strip.show();

  handleAnimationState(nowMillis);
}

void handleAnimationState(unsigned long nowMillis) {
  if (!animating) {
    return;
  }
  
  byte nextAnimationState = (animationState == ANIMATION_STATE_FADING) ?
    ANIMATION_STATE_GROWING :
    ANIMATION_STATE_FADING;
  
  if (nowMillis >= animationEndMillis) {
    animationState = nextAnimationState;
    recordAnimationTiming(nowMillis);
  }
}

//
// TODO: pull the RGB out of this color, and scale appropriately based on the animation progress
//
uint32_t calculateAnimatedColor(unsigned long nowMillis, uint32_t baseColor) {
  if (!animating) {
    return baseColor;
  }
  // It's likely that we're past the planned end time of the animation, so clamp the
  // time to that end value for the purposes of scaling the color value.
  unsigned long animationTime = constrain(nowMillis, animationStartMillis, animationEndMillis);
  
  uint8_t r = baseColor >> 16; 
  uint8_t g = baseColor >> 8;
  uint8_t b = baseColor;
  
  switch (animationState) {
    case ANIMATION_STATE_GROWING:
      r = map(animationTime, animationStartMillis, animationEndMillis, 0, r);
      g = map(animationTime, animationStartMillis, animationEndMillis, 0, g);
      b = map(animationTime, animationStartMillis, animationEndMillis, 0, b);
      break;
    case ANIMATION_STATE_FADING:
      r = map(animationTime, animationStartMillis, animationEndMillis, r, 0);
      g = map(animationTime, animationStartMillis, animationEndMillis, g, 0);
      b = map(animationTime, animationStartMillis, animationEndMillis, b, 0);
      break;
  }

  return strip.Color(r, g, b);
}

void recordAnimationTiming(unsigned long nowMillis) {
  animationStartMillis = nowMillis;
  animationEndMillis = animationStartMillis + 500L;

  if (animationEndMillis < animationStartMillis) {
    // We overflowed the long value. Wait another loop to get a usable calculation
    animationStartMillis = -1;
    animationEndMillis = -1;
  }
}

void handleEncoderButton() {
  // The switch is active low
  if (bit_is_clear(PORT_D_INPUTS, PIN_ENCODER_SWITCH)) {
    if (encSwitchWasPressed == 0) {
      // only on initial press, so the keystroke is not repeated while the button is held down
      delay(5); // debounce delay
    }
    encSwitchWasPressed = 1;
  } else {
    if (encSwitchWasPressed != 0) {
      delay(5); // debounce delay
      toggleMode();
    }
    encSwitchWasPressed = 0;
  }
}

void adjustBrightness() {
  static int prevBrightness = 0;
  brightness = map(analogRead(PIN_PHOTORESISTOR), 0, 1023, 1, 255);

  int deltaBrightness = abs(brightness - prevBrightness);
  // Helps protect against small differences in the readings causing
  // lots of imperceptible adjustments
  if (deltaBrightness > 2) {
    strip.setBrightness(brightness);
    prevBrightness = brightness;
    updatedAppearance = true;
  }
}

void setDigitAt(unsigned int index, unsigned int value, uint32_t color) {
  unsigned int i;
  unsigned int offset = index * SEGMENTS_PER_DIGIT;
  
  // The two pixels for the colon are between the first two and last two digits.
  // When addressing pixels for the last two digits, account for the two pixels
  // that aren't a part of any digit.
  if (index >= 2) {
    offset += 2;
  }
  
  for (i = 0; i < SEGMENTS_PER_DIGIT; i++) {
    unsigned int segmentState = DIGIT_SEGMENTS[value][i];
    if (segmentState == 0) {
      strip.setPixelColor(offset + i, 0);
    } else {
      strip.setPixelColor(offset + i, color);
    }
  }
}

void toggleMode() {
  mode = ++mode % MODE_COUNT;
  switch (mode) {
    case NOOP_MODE:
      Serial.println("No-op mode");
      animating = false;
      animationState = ANIMATION_STATE_NONE;
      break;
    case SET_HOUR_MODE:
      Serial.println("Set hour mode");
      animating = true;
      break;
    case SET_MINUTE_MODE:
      Serial.println("Set minute mode");
      animating = true;
      break;
  }
}

// note: for better performance, the code will use
// direct port access techniques
// http://www.arduino.cc/en/Reference/PortManipulation
int8_t getEncoderRotation() {
  static uint8_t encFlags = 0;
  uint8_t encCurPos = 0;
  int8_t encAction = 0; // 1 or -1 if moved, sign is direction

  // read in the encoder state first
  if (bit_is_clear(PORT_D_INPUTS, PIN_ENCODER_A)) {
    encCurPos |= (1 << 0);
  }
  if (bit_is_clear(PORT_D_INPUTS, PIN_ENCODER_B)) {
    encCurPos |= (1 << 1);
  }

  // if any rotation at all
  if (encCurPos != encPrevPos) {
    if (encPrevPos == 0x00) {
      // this is the first edge
      if (encCurPos == 0x01) {
        encFlags |= (1 << 0);
      } else if (encCurPos == 0x02) {
        encFlags |= (1 << 1);
      }
    }

    if (encCurPos == 0x03) {
      // this is when the encoder is in the middle of a "step"
      encFlags |= (1 << 4);
    } else if (encCurPos == 0x00) {
      // this is the final edge
      if (encPrevPos == 0x02) {
        encFlags |= (1 << 2);
      } else if (encPrevPos == 0x01) {
        encFlags |= (1 << 3);
      }

      uint8_t flag0 = bit_is_set(encFlags, 0);
      uint8_t flag1 = bit_is_set(encFlags, 1);
      uint8_t flag2 = bit_is_set(encFlags, 2);
      uint8_t flag3 = bit_is_set(encFlags, 3);
      uint8_t flag4 = bit_is_set(encFlags, 4);

      // check the first and last edge
      // or maybe one edge is missing, if missing then require the middle state
      // this will reject bounces and false movements
      if (flag0 && (flag2 || flag4)) {
        encAction = 1;
      } else if (flag2 && (flag0 || flag4)) {
        encAction = 1;
      } else if (flag1 && (flag3 || flag4)) {
        encAction = -1;
      } else if (flag3 && (flag1 || flag4)) {
        encAction = -1;
      }

      encFlags = 0;
    }
  }

  encPrevPos = encCurPos;
  return encAction;
}

// Input a value 0 to 255 to get a color value.
// The colours are a transition r - g - b - back to r.
uint32_t wheel(byte wheelPos) {
  wheelPos = 255 - wheelPos;
  if (wheelPos < 85) {
    return strip.Color(255 - wheelPos * 3, 0, wheelPos * 3);
  }
  if (wheelPos < 170) {
    wheelPos -= 85;
    return strip.Color(0, wheelPos * 3, 255 - wheelPos * 3);
  }
  wheelPos -= 170;
  return strip.Color(wheelPos * 3, 255 - wheelPos * 3, 0);
}
